/****************************************************************
Copyright (c) 1998 Sandro Sigala <ssigala@globalnet.it>.
All rights reserved.

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name of the author not be used in
advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

The author disclaim all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall the author be liable for any
special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
****************************************************************/

#include "config.h"

#include <mediatool.h>
#include <kaudio.h>
#include <kapp.h>
#include <kmsgbox.h>
#include <qaccel.h>
#include <qbitmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qtimer.h>
#include <stdlib.h>
#ifdef HAVE_SLEEP
#include <unistd.h>
#endif

#include "kzacman.h"
#include "gamewidget.h"
#include "scoredialog.h"
#include "newscoredialog.h"

GameWidget::GameWidget(QWidget *parent, const char *name)
        : QWidget(parent, name)
{
	audio = new KAudio();
	monsters.setAutoDelete(true);
	in_game = false;
	map = 0;

	QAccel *accel = new QAccel(this);
	accel->connectItem(accel->insertItem(Key_Up), this, SLOT(goUp()));
	accel->connectItem(accel->insertItem(Key_Down), this, SLOT(goDown()));
	accel->connectItem(accel->insertItem(Key_Left), this, SLOT(goLeft()));
	accel->connectItem(accel->insertItem(Key_Right), this, SLOT(goRight()));

	loadSprites();

	timerSecs = startTimer(1000);
	timerMonsters = startTimer(300);
}

GameWidget::~GameWidget()
{
	delete audio;
	delete [] map;
	delete [] sprites;
}

void GameWidget::playSound(Sound s)
{
	if (!do_sounds || audio->serverStatus() != 0)
		return;

	QString name;
	switch (s) {
	case Sound_Die:
		name = "die.wav";
		break;
	case Sound_Got_Point:
		name = "point.wav";
		break;
	case Sound_Got_Potion:
		name = "potion.wav";
		break;
	case Sound_Killed_Monster:
		name = "monster.wav";
		break;
	case Sound_Saw_Monster:
		name = "saw.wav";
		break;
	}

	QString fname = kapp->kde_datadir() + "/kzacman/wav/" + name;
//	audio->stop();
	audio->play(fname);
//	audio->sync();
}

void GameWidget::loadSprites()
{
	sprites = new QPixmap[num_sprites];
	loadSprite(Sprite_Back1, "back1.bmp");
	loadSprite(Sprite_Back2, "back2.bmp");
	loadSprite(Sprite_Back3, "back3.bmp");
	loadSprite(Sprite_Back4, "back4.bmp");
	loadSprite(Sprite_Back5, "back5.bmp");
	loadSprite(Sprite_Back6, "back6.bmp");
	loadSprite(Sprite_Wall1, "wall1.bmp");
	loadSprite(Sprite_Wall2, "wall2.bmp");
	loadSprite(Sprite_Wall3, "wall3.bmp");
	loadSprite(Sprite_Wall4, "wall4.bmp");
	loadSprite(Sprite_Wall5, "wall5.bmp");
	loadSprite(Sprite_Wall6, "wall6.bmp");
	loadMaskedSprite(Sprite_Zacman_Up, "zacman-up.bmp", "zacman-up-mask.bmp");
	loadMaskedSprite(Sprite_Zacman_Down, "zacman-down.bmp", "zacman-down-mask.bmp");
	loadMaskedSprite(Sprite_Zacman_Left, "zacman-left.bmp", "zacman-left-mask.bmp");
	loadMaskedSprite(Sprite_Zacman_Right, "zacman-right.bmp", "zacman-right-mask.bmp");
	loadMaskedSprite(Sprite_Zacman_Potion_Up, "zacman-potion-up.bmp", "zacman-up-mask.bmp");
	loadMaskedSprite(Sprite_Zacman_Potion_Down, "zacman-potion-down.bmp", "zacman-down-mask.bmp");
	loadMaskedSprite(Sprite_Zacman_Potion_Left, "zacman-potion-left.bmp", "zacman-left-mask.bmp");
	loadMaskedSprite(Sprite_Zacman_Potion_Right, "zacman-potion-right.bmp", "zacman-right-mask.bmp");
	loadMaskedSprite(Sprite_Monster_Up, "monster-up.bmp", "monster-up-mask.bmp");
	loadMaskedSprite(Sprite_Monster_Down, "monster-down.bmp", "monster-down-mask.bmp");
	loadMaskedSprite(Sprite_Monster_Left, "monster-left.bmp", "monster-left-mask.bmp");
	loadMaskedSprite(Sprite_Monster_Right, "monster-right.bmp", "monster-right-mask.bmp");
	loadMaskedSprite(Sprite_Point, "point.bmp", "point-mask.bmp");
	loadMaskedSprite(Sprite_Potion, "potion.bmp", "potion-mask.bmp");
	loadMaskedSprite(Sprite_Arrow_Up, "arrow-up.bmp", "arrow-up-mask.bmp");
	loadMaskedSprite(Sprite_Arrow_Down, "arrow-down.bmp", "arrow-down-mask.bmp");
	loadMaskedSprite(Sprite_Balloon_Ouch, "balloon-ouch.bmp", "balloon-ouch-mask.bmp");
	loadMaskedSprite(Sprite_Balloon_Chomp, "balloon-chomp.bmp", "balloon-chomp-mask.bmp");
	loadMaskedSprite(Sprite_Balloon_Exit, "balloon-exit.bmp", "balloon-exit-mask.bmp");
}

void GameWidget::loadSprite(Sprite spr, const char *path)
{
	if (!sprites[spr].load(kapp->kde_datadir() + "/kzacman/data/" + path))
		fatal("Cannot open data files.\nHave you correctly installed KZacman?");
}

void GameWidget::loadMaskedSprite(Sprite spr, char *path1, const char *path2)
{
	QBitmap bmp;
	if (!sprites[spr].load(kapp->kde_datadir() + "/kzacman/data/" + path1))
		fatal("Cannot open data files.\nHave you correctly installed KZacman?");
	if (!bmp.load(kapp->kde_datadir() + "/kzacman/data/" + path2))
		fatal("Cannot open data files.\nHave you correctly installed KZacman?");
	sprites[spr].setMask(bmp);
}

void GameWidget::newGame(Skill skill)
{
	game_skill = skill;
	in_pause = false;
	show_balloon = false;
	xpos = ypos = 0;
	dir = South;
	sprite_wall = (Sprite)(Sprite_Wall1 + rand() % num_walls);
	sprite_back = (Sprite)(Sprite_Back1 + rand() % num_backs);
	deleteMonsters();

	int map_prop = 6;
	switch (skill) {
	case Skill_Beginner:
		map_prop = 6;
		break;
	case Skill_Intermediate:
		map_prop = 9;
		break;
	case Skill_Advanced:
		map_prop = 12;
		break;
	case Skill_Expert:
		map_prop = 15;
		break;
	}
	map_height = map_width = (4 * map_prop) + 1 - 4;
	if (map) delete [] map;
	map = new Sprite[map_width * map_height];

	for (int y = 0; y < map_height; ++y)
		for (int x = 0; x < map_width; ++x)
			ref(x, y) = sprite_wall;
	for (int y = 0; y < map_height; ++y)
		ref(0, y) = ref(map_width - 1, y) = sprite_back;
	for (int x = 0; x < map_width; ++x)
		ref(x, 0) = ref(x, map_height - 1) = sprite_back;

	// Compute the maze
	createMap(map_width / 2, map_height / 2);

	for (int x = 0; x < map_width; ++x)
		ref(x, 0) = ref(x, map_height - 1) = sprite_wall;
	for (int y = 0; y < map_height; ++y)
		ref(0, y) = ref(map_width - 1, y) = sprite_wall;

	// Clear the start and end path
	int pos[][2] = {{0, 0}, {0, 1}, {0, 2}, {1, 2}};
	for (int i = 0; i < 4; ++i)
		ref(pos[i][0], pos[i][1]) =
			ref(map_width - 1 - pos[i][0],
			    map_height - 1 - pos[i][1]) = sprite_back;

	// Place randomly the points
	num_points = 25 * (skill + 1);
	for (int i = 0; i < num_points; ) {
		int x = rand() % map_width;
		int y = rand() % map_height;
		if (!(x == 0 && y == 0) && ref(x, y) == sprite_back) {
			ref(x, y) = Sprite_Point;
			++i;
		}
	}

	// Remove randomly some walls
	for (int i = 0; i < (map_width * map_height) / 40; ) {
		int x = rand() % map_width;
		int y = rand() % map_height;
		if (x >= 2 && y >= 2
		    && x <= map_width - 3 && y <= map_height - 3
		    && ref(x, y) == sprite_wall
		    && ((ref(x-1, y) == sprite_back && ref(x, y-1) == sprite_wall
			 && ref(x+1, y) == sprite_back && ref(x, y+1) == sprite_wall)
			|| (ref(x-1, y) == sprite_wall && ref(x, y-1) == sprite_back
			    && ref(x+1, y) == sprite_wall && ref(x, y+1) == sprite_back))) {
			ref(x, y) = sprite_back;
			++i;
		}
	}

	// Place randomly the monsters
	for (int i = 0; i < (skill + 1) * 4; ) {
		int x = rand() % map_width;
		int y = rand() % map_height;
		if (x >= 10 && y >= 10 && ref(x, y) == sprite_back) {
			addMonster(x, y);
			++i;
		}
	}

	// Place randomly the potions
	for (int i = 0; i < (skill + 1) * 2; ) {
		int x = rand() % map_width;
		int y = rand() % map_height;
		if (x >= 3 && y >= 3 && ref(x, y) == sprite_back) {
			ref(x, y) = Sprite_Potion;
			++i;
		}
	}

	num_secs = 0;
	num_potion_secs = 0;
	in_game = true;
	repaint(false);
	emit changedStats(num_points, num_secs, num_potion_secs);
}

void GameWidget::addMonster(int x, int y)
{
	monsters.append(new Monster(x, y, Sprite_Monster_Up));
}

Monster *GameWidget::findMonster(int x, int y)
{
	Monster *m;
	for (m = monsters.first(); m != 0; m = monsters.next())
		if (m->xpos == x && m->ypos == y)
			return m;
	return 0;
}

void GameWidget::killMonster(Monster *mp)
{
	for (Monster *m = monsters.first(); m != 0; m = monsters.next())
		if (m == mp) {
			monsters.remove();
			playSound(Sound_Killed_Monster);
			showBalloon(Sprite_Balloon_Chomp, 0);
		}
}

void GameWidget::deleteMonsters()
{
	Monster *m;
	while ((m = monsters.first()) != 0)
		monsters.remove();

}

// Try to look around for Zacman
bool GameWidget::lookMonster(Monster *m)
{
	if (m->ypos == ypos) {
		for (int x = m->xpos - 1; x >= 0; --x)
			if (xy(x, m->ypos) == sprite_wall)
				break;
			else if (xpos == x) {
				m->dir = West;
				return true;
			}
		for (int x = m->xpos + 1; x <= map_width - 1; ++x)
			if (xy(x, m->ypos) == sprite_wall)
				break;
			else if (xpos == x) {
				m->dir = East;
				return true;
			}
	}
	if (m->xpos == xpos) {
		for (int y = m->ypos - 1; y >= 0; --y)
			if (xy(m->xpos, y) == sprite_wall)
				break;
			else if (ypos == y) {
				m->dir = North;
				return true;
			}
		for (int y = m->ypos + 1; y <= map_height - 1; ++y)
			if (xy(m->xpos, y) == sprite_wall)
				break;
			else if (ypos == y) {
				m->dir = South;
				return true;
			}
	}

	return false;
}

Direction GameWidget::monsterFindDirection(Monster *m, Direction exclude)
{
	if (exclude != South && xy(m->xpos, m->ypos + 1) != sprite_wall)
		return South;
	else if (exclude != West && xy(m->xpos - 1, m->ypos) != sprite_wall)
		return West;
	else if (exclude != North && xy(m->xpos, m->ypos - 1) != sprite_wall)
		return North;
	else if (exclude != East && xy(m->xpos + 1, m->ypos) != sprite_wall)
		return East;

	switch (exclude) {
	case North: return South;
	case South: return North;
	case West: return East;
	case East: return West;
	}

	return North; // Never happens
}

// Update the monsters' status
void GameWidget::updateMonsters()
{
	Monster *m;
	for (m = monsters.first(); m != 0; m = monsters.next()) {
		Sprite s;
		while ((s = (Sprite)(Sprite_Monster_Up + rand() % 4)) == m->sprite)
			;
		m->sprite = s;
		if (lookMonster(m)) {
			if (num_potion_secs > 0) // Invert direction
				switch (m->dir) {
				case North:
					m->dir = monsterFindDirection(m, North);
					break;
				case South:
					m->dir = monsterFindDirection(m, South);
					break;
				case West:
					m->dir = monsterFindDirection(m, West);
					break;
				case East:
					m->dir = monsterFindDirection(m, East);
				}
			else {
				if (!m->saw) {
					playSound(Sound_Saw_Monster);
					m->saw = true;
				}
			}
		} else
			m->saw = false;
		if (m->dir == North && xy(m->xpos, m->ypos - 1) != sprite_wall) {
			if (xpos == m->xpos && ypos == m->ypos - 1) {
				if (num_potion_secs > 0)
					killMonster(m);
				else
					die();
				return;
			}
			--m->ypos;
		} else if (m->dir == South && xy(m->xpos, m->ypos + 1) != sprite_wall) {
			if (xpos == m->xpos && ypos == m->ypos + 1) {
				if (num_potion_secs > 0)
					killMonster(m);
				else
					die();
				return;
			}
			++m->ypos;
		} else if (m->dir == East && xy(m->xpos + 1, m->ypos) != sprite_wall) {
			if (xpos == m->xpos + 1 && ypos == m->ypos) {
				if (num_potion_secs > 0)
					killMonster(m);
				else
					die();
				return;
			}
			++m->xpos;
		} else if (m->dir == West && xy(m->xpos - 1, m->ypos) != sprite_wall) {
			if (xpos == m->xpos - 1 && ypos == m->ypos) {
				if (num_potion_secs > 0)
					killMonster(m);
				else
					die();
				return;
			}
			--m->xpos;
		} else {
			Direction d = (Direction)(rand() % 4);
			m->dir = d;
		}
	}
	repaint(false);
}

void GameWidget::createMap(int x, int y)
{
	int dir = 0;

	while (ref(x, y + 2) != sprite_back || ref(x, y - 2) != sprite_back
	    || ref(x + 2, y) != sprite_back || ref(x - 2, y) != sprite_back) {
		bool found = false;
		while (!found) {
			dir = rand() % 4;
			switch (dir) {
			case 0:
				if (ref(x, y + 2) == sprite_wall)
					found = true;
				break;
			case 1:
				if (ref(x, y - 2) == sprite_wall)
					found = true;
				break;
			case 2:
				if (ref(x + 2, y) == sprite_wall)
					found = true;
				break;
			case 3:
				if (ref(x - 2, y) == sprite_wall)
					found = true;
			}
		}
		ref(x, y) = sprite_back;
		switch (dir) {
		case 0:
			ref(x, y + 2) = ref(x, y + 1) = sprite_back;
			createMap(x, y + 2);
			break;
		case 1:
			ref(x, y - 2) = ref(x, y - 1) = sprite_back;
			createMap(x, y - 2);
			break;
		case 2:
			ref(x + 2, y) = ref(x + 1, y) = sprite_back;
			createMap(x + 2, y);
			break;
		case 3:
			ref(x - 2, y) = ref(x - 1, y) = sprite_back;
			createMap(x - 2, y);
		}
	}
}

void GameWidget::checkPosition(int x, int y)
{
	if (xy(x, y) == Sprite_Point) {
		playSound(Sound_Got_Point);
		ref(x, y) = sprite_back;
		--num_points;
	} else if (xy(x, y) == Sprite_Potion) {
		playSound(Sound_Got_Potion);
		ref(x, y) = sprite_back;
		num_potion_secs += 10;
	}
	emit changedStats(num_points, num_secs, num_potion_secs);
	if (x == map_width - 1 && y == map_height - 1)
		if (!num_points) {
			in_game = false;
			KMsgBox::message(this, "KZacman", "Game finished!");
			KConfig *config = kapp->getConfig();
			config->setGroup("High Score");
			QString s, time;
			switch (game_skill) {
			case Skill_Beginner: s = "Beginner"; break;
			case Skill_Intermediate: s = "Intermediate"; break;
			case Skill_Advanced: s = "Advanced"; break;
			case Skill_Expert: s = "Expert";
			}
			time = config->readEntry(s + "Time", "9999");
			if (num_secs < time.toInt()) {
				NewScoreDialog d(this);
				if (d.exec()) {
					time.setNum(num_secs);
					config->writeEntry(s + "Time", time);
					config->writeEntry(s + "Name", d.name());

					ScoreDialog h(this);
					h.exec();
				}
			}
		} else
			showBalloon(Sprite_Balloon_Exit, 0);
}

void GameWidget::goUp()
{
	if (!in_game || in_pause) return;
	if (show_balloon)
		show_balloon = false;

	Monster *m;
	if ((m = findMonster(xpos, ypos - 1)) != 0)
		if (num_potion_secs > 0)
			killMonster(m);
		else {
			die();
			return;
		}
	if (xy(xpos, ypos - 1) != sprite_wall)
		--ypos;
	checkPosition(xpos, ypos);
	dir = North;
	repaint(false);
}

void GameWidget::goDown()
{
	if (!in_game || in_pause) return;
	if (show_balloon)
		show_balloon = false;

	Monster *m;
	if ((m = findMonster(xpos, ypos + 1)) != 0)
		if (num_potion_secs > 0)
			killMonster(m);
		else {
			die();
			return;
		}
	if (xy(xpos, ypos + 1) != sprite_wall)
		++ypos;
	checkPosition(xpos, ypos);
	dir = South;
	repaint(false);
}

void GameWidget::goLeft()
{
	if (!in_game || in_pause) return;
	if (show_balloon)
		show_balloon = false;

	Monster *m;
	if ((m = findMonster(xpos - 1, ypos)) != 0)
		if (num_potion_secs > 0)
			killMonster(m);
		else {
			die();
			return;
		}
	if (xy(xpos - 1, ypos) != sprite_wall)
		--xpos;
	checkPosition(xpos, ypos);
	dir = West;
	repaint(false);
}

void GameWidget::goRight()
{
	if (!in_game || in_pause) return;
	if (show_balloon)
		show_balloon = false;

	Monster *m;
	if ((m = findMonster(xpos + 1, ypos)) != 0)
		if (num_potion_secs > 0)
			killMonster(m);
		else {
			die();
			return;
		}
	if (xy(xpos + 1, ypos) != sprite_wall)
		++xpos;
	checkPosition(xpos, ypos);
	dir = East;
	repaint(false);
}

void GameWidget::timerEvent(QTimerEvent *e)
{
	if (!in_game || in_pause)
		return;

	if (e->timerId() == timerSecs) {
		++num_secs;
		if (num_potion_secs > 0) {
			--num_potion_secs;
			if (!num_potion_secs)
				playSound(Sound_Got_Potion);
		}
		emit changedStats(num_points, num_secs, num_potion_secs);
	} else if (e->timerId() == timerMonsters)
		updateMonsters();
}

void GameWidget::die()
{
	playSound(Sound_Die);
	showBalloon(Sprite_Balloon_Ouch, 2);
	in_game = false;
	repaint(false);
	KMsgBox::message(this, "KZacman", "Game Over");
}

void GameWidget::showBalloon(Sprite s, int secs)
{
	show_balloon = true;
	balloon = s;
	if (secs > 0) {
		repaint(false);
		repaint(false);	// Needed twice, don't know why
#ifdef HAVE_SLEEP
		sleep(secs);
#endif
		show_balloon = false;
	}
}

void GameWidget::paintEvent(QPaintEvent *)
{
	if (!in_game) {
		QPixmap bg(kapp->kde_datadir() + "/kzacman/data/bg.bmp");
		for (int y = 0; y < height() / bg.height() + 1; ++y)
			for (int x = 0; x < width() / bg.width() + 1; ++x)
				bitBlt(this, x*bg.width(), y*bg.height(), &bg);
		emit noStats();
		return;
	}

	QPixmap pm(size());
	QPainter p;
	p.begin(&pm);

	for (int y = 0; y < height() / sprite_height; y++)
		for (int x = 0; x < width() / sprite_width; x++) {
			int mx = x + xpos - width() / sprite_width / 2;
			int my = y + ypos - height() / sprite_height / 2;
			Sprite s = xy(mx, my);
			if (s == Sprite_Point || s == Sprite_Potion)
				p.drawPixmap(x*32, y*32, sprites[sprite_back]);
			p.drawPixmap(x*32, y*32, sprites[s]);

			Monster *m;
			if ((m = findMonster(mx, my)) != 0)
				p.drawPixmap(x*32, y*32, sprites[m->sprite]);
		}

	int scr_dwidth = (width() / 32 - 1) / 2;
	if ((width() / 32) % 2 == 0)
		++scr_dwidth;
	int scr_dheight = (height() / 32 - 1) / 2;
	if ((height() / 32) % 2 == 0)
		++scr_dheight;
	if (xpos <= scr_dwidth && ypos < scr_dheight)
		p.drawPixmap((scr_dwidth - xpos) * 32,
			     (scr_dheight - ypos - 1) * 32,
			     sprites[Sprite_Arrow_Down]);
	if (xpos >= map_width - scr_dwidth - 1
	    && ypos > map_height - scr_dheight - 1)
		p.drawPixmap((scr_dwidth - xpos + map_width - 1) * 32,
			     (scr_dheight - ypos + map_height) * 32,
			     sprites[Sprite_Arrow_Up]);

	Sprite s = Sprite_Zacman_Up; // Avoid uninitialized warning
        if (num_potion_secs > 0)
		switch (dir) {
		case North: s = Sprite_Zacman_Potion_Up; break;
		case South: s = Sprite_Zacman_Potion_Down; break;
		case West: s = Sprite_Zacman_Potion_Left; break;
		case East: s = Sprite_Zacman_Potion_Right;
		}
	else
		switch (dir) {
		case North: s = Sprite_Zacman_Up; break;
		case South: s = Sprite_Zacman_Down; break;
		case West: s = Sprite_Zacman_Left; break;
		case East: s = Sprite_Zacman_Right;
		}

	p.drawPixmap(scr_dwidth * 32, scr_dheight * 32, sprites[s]);

	if (show_balloon)
		p.drawPixmap(scr_dwidth * 32 - sprites[balloon].width(),
			     scr_dheight * 32 - sprites[balloon].height(),
			     sprites[balloon]);

	if (in_pause) {
		QPixmap pause(kapp->kde_datadir() + "/kzacman/data/pause.bmp");
		p.drawPixmap((width()-pause.width())/2, 
			     (height()-pause.height())/2, pause);
	}

	p.end();
	bitBlt(this, 0, 0, &pm);
}
